/*
 * cntlr_map.c - implements MAP2 CMDUs handling
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <pthread.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <uci.h>

#include <easy/easy.h>
#include <easy/utils.h>

#include <timer_impl.h>
#include <cmdu.h>
#include <i1905_wsc.h>
#include <cntlrsync.h>
#include <1905_tlvs.h>
#include <map2.h>
#include <map_module.h>

#include "utils/utils.h"
#include "utils/debug.h"
#include "utils/liblist.h"
#include "config.h"
#include "cntlr.h"
#include "cntlr_ubus.h"

#include "cntlr_map_debug.h"
#include "cntlr_cmdu.h"
#include "cntlr_tlv.h"
#include "cmdu_validate.h"

#define TIMESTAMP_MAX_LEN 256

typedef int (*map_cmdu_handler_t)(void *cntlr, struct cmdu_buff *cmdu);

struct map_cmdu_calltable_t {
	map_cmdu_handler_t handle;
	map_cmdu_handler_t debug;
};

int handle_topology_notification(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);
	return 0;
}

static void _cntlr_update_steer_params(struct controller *c, struct opclass *op)
{
	int i, j, k;
	bool found;
	struct opclass_entry *oe;
	struct steering *sp;

	for (i = 0; i < op->opclass_entry_num; i++) {
		oe = &op->opclass_entry[i];
		for (j = 0; j < oe->channels_num; j++) {
			sp = &c->steer_params;
			found = false;
			for (k = 0; k < sp->channels_num; k++) {
				if(sp->channels[k] == oe->channels[j].channel) {
					found = true;
					break;
				}
			}

			if (!found) {
				sp->channels[sp->channels_num] = oe->channels[j].channel;
				sp->channels_num++;
			}
		}
	}
}

static void cntlr_update_steer_params(struct controller *c, struct sta *s)
{
    struct netif_iface *p = NULL;  /* fh anf bk iface */
    struct netif_radio *r = NULL;
    struct node *n = NULL;

    list_for_each_entry(n, &c->nodelist, list) {
        list_for_each_entry(r, &n->radiolist, list) {
            list_for_each_entry(p, &r->iflist, list) {
				if (p->ssid && s->fh && s->fh->ssid
				    && !memcmp(p->ssid, s->fh->ssid, 33))
					_cntlr_update_steer_params(c, &r->cur_opclass);
            }
        }
    }
}

int handle_topology_response(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);

	struct controller *c = (struct controller *) cntlr;
	struct tlv *tv[11][16] = {0};
	struct node *n;
#ifdef EASYMESH_VENDOR_EXT
	int num = 0;
#endif

	n = find_node_by_mac(c, cmdu->origin);
	if (!n)
		return -1;

	if (!validate_topology_response(cmdu, tv)) {
		dbg("cmdu validation: [TOPOLOGY_RESPONSE] failed\n");
		return -1;
	}

	if (tv[7][0]) {
		struct tlv_ap_oper_bss *tlv;
		uint8_t *tv_data;
		int i, offset = 0;

		tlv = (struct tlv_ap_oper_bss *)tv[7][0]->data;
		if (!tlv)
			return -1;

		tv_data = (uint8_t *)tlv;

		offset += 1; /* num_radio */

		for (i = 0; i < tlv->num_radio; i++) {
			uint8_t num_bss = 0;
			struct netif_radio *r = NULL;
			struct netif_iface *p = NULL;
			int j;

			r = cntlr_node_add_radio(c, n, &tv_data[offset]);
			if (!r)
				return -1; /* FIXME: continue + proper offset shift */

			offset += 6; /* macaddr */

			/* disable all prior stored fh/bh interfaces */
			list_for_each_entry(p, &r->iflist, list) {
				if (p->type == NETIF_BSTA)
					continue;
				p->active = false;
				p->type = NETIF_FHBSS;
			}

			memcpy(&num_bss, &tv_data[offset], 1);

			offset += 1; /* num_bss */

			for (j = 0; j < num_bss; j++) {
				uint8_t ssidlen = 0, len;
				struct netif_iface *fh;

				fh = cntlr_radio_add_interface(c, r, &tv_data[offset]);
				if (!fh)
					return -1; /* FIXME: continue + proper offset shift */

				offset += 6; /* macaddr */

				memcpy(&ssidlen, &tv_data[offset], 1);

				offset += 1; /* ssidlen */

				memset(fh->ssid, 0, sizeof(fh->ssid));
				len = (ssidlen > sizeof(fh->ssid) - 1
						? sizeof(fh->ssid) - 1 : ssidlen);
				memcpy(fh->ssid, &tv_data[offset], len);

				offset += ssidlen; /* ssid */
			}
		}
	}
#ifdef EASYMESH_VENDOR_EXT
	/* vendor tlv containing backhaul interfaces only */
	while (tv[10][num]) {
		struct tlv_vendor_bbss *tlv;
		int i, offset = 0;
		uint8_t oui2[3];  /* TODO: use the same vendor oui-type */

		tlv = (struct tlv_vendor_bbss *)tv[10][num]->data;
		if (!tlv)
			return -1;

		/* oui (3 bytes) */
		memcpy(oui2, EASYMESH_VENDOR_EXT_OUI, 3);
		oui2[2]++;

		if (memcmp(tlv->oui, oui2, 3)) {
			num++;
			continue;
		}

		offset += 3; /* oui */

		offset += 1; /* num_radios */

		for (i = 0; i < tlv->num_radios; i++) {
			uint8_t num_bss = 0;
			struct netif_radio *r;
			int j;

			r = cntlr_node_add_radio(c, n,
					(uint8_t *)&tv[10][num]->data[offset]);
			if (!r)
				return -1; /* FIXME: continue + proper offset shift */

			offset += 6; /* macaddr */

			memcpy(&num_bss, &tv[10][num]->data[offset], 1);

			offset += 1; /* num_bss */

			for (j = 0; j < num_bss; j++) {
				struct netif_iface *fh;

				fh = cntlr_radio_add_interface(c, r,
						(uint8_t *)&tv[10][num]->data[offset]);
				if (!fh)
					return -1; /* FIXME: continue + proper offset shift */

				fh->type = NETIF_BKBSS;

				offset += 6; /* macaddr */
			}
		}

		num++;
	}
#endif /*EASYMESH_VENDOR_EXT*/

	if (tv[8][0]) {
		struct tlv_assoc_client *tlv;
		uint8_t *tv_data;
		int i, offset = 0;

		tlv = (struct tlv_assoc_client *)tv[8][0]->data;
		if (!tlv)
			return -1;

		tv_data = (uint8_t *)tlv;

		offset += 1; /* num_bss */

		for (i = 0; i < tlv->num_bss; i++) {
			struct netif_iface *fh;
			uint8_t bssid[6] = {0};
			uint16_t num_client = 0;
			int j;

			memcpy(bssid, &tv_data[offset], 6);

			offset += 6; /* radio_id */

			num_client = BUF_GET_BE16(tv_data[offset]);

			offset += 2; /* num_client */

			fh = cntlr_iterate_fbss(c, bssid);
			if (!fh) {
				/* num_client * (macaddr + conntime) */
				offset += (num_client * 8);

				continue;
			}

			for (j = 0; j < num_client; j++) {
				uint8_t macaddr[6] = {0};
				uint16_t conntime;
				struct sta *s;
				struct netif_iface *bsta;

				memcpy(macaddr, &tv_data[offset], 6);

				offset += 6; /* macaddr */

				conntime = BUF_GET_BE16(tv_data[offset]);

				offset += 2; /* conntime */

				s = cntlr_add_sta(c, macaddr);
				if (!s)
					continue;

				memcpy(s->bssid, bssid, 6);
				s->fh = fh;
				s->conntime = conntime;

				bsta = cntlr_iterate_fbss(c, macaddr);
				if (bsta) {
					/* The client is an agent and the iface is a bSTA */
					memcpy(bsta->upstream_bssid, bssid, 6);
					bsta->type = NETIF_BSTA;
					bsta->active = true;
					s->type = IEEE1905;
					s->agent = bsta->agent;
					//s->failed_steer_attempts = 0;
					//timestamp_reset(s->stats.last_steer_time);
				}

				/* Get STA data for steering and beacon requests */
				cntlr_update_steer_params(c, s);
			}
		}
	}

	/* Check opclass preferency age */
	if (cntlr_node_opclass_expired(n)) {
		trace("node " MACFMT " opclass expired\n", MAC2STR(n->alid));
		cntrl_send_channel_preference_query(c, n->alid);
	}

	return 0;
}

static inline uint16_t c_cmdu_expect_response(uint16_t req_type)
{
	switch (req_type) {
		case CMDU_AP_CAPABILITY_QUERY:
			return CMDU_AP_CAPABILITY_REPORT;
		case CMDU_POLICY_CONFIG_REQ:
			return CMDU_1905_ACK;
		case CMDU_CHANNEL_PREFERENCE_QUERY:
			return CMDU_CHANNEL_PREFERENCE_REPORT;
		case CMDU_CHANNEL_SELECTION_REQ:
			return CMDU_CHANNEL_SELECTION_RESPONSE;
		case CMDU_OPERATING_CHANNEL_REPORT:
			return CMDU_1905_ACK;
		case CMDU_CLIENT_CAPABILITY_QUERY:
			return CMDU_CLIENT_CAPABILITY_REPORT;
		case CMDU_AP_METRICS_QUERY:
			return CMDU_AP_METRICS_RESPONSE;
		case CMDU_ASSOC_STA_LINK_METRICS_QUERY:
			return CMDU_ASSOC_STA_LINK_METRICS_RESPONSE;
		case CMDU_UNASSOC_STA_LINK_METRIC_QUERY:
			return CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE;
		case CMDU_BEACON_METRICS_QUERY:
			return CMDU_BEACON_METRICS_RESPONSE;
		case CMDU_COMBINED_INFRA_METRICS:
			return CMDU_1905_ACK;
		case CMDU_CLIENT_STEERING_REQUEST:
			// FIX THIS: we need ACK ?
			return CMDU_CLIENT_STEERING_BTM_REPORT;
		case CMDU_CLIENT_ASSOC_CONTROL_REQUEST:
			return CMDU_1905_ACK;
		case CMDU_STEERING_COMPLETED:
			return CMDU_1905_ACK;
		case CMDU_HIGHER_LAYER_DATA:
			return CMDU_1905_ACK;
		case CMDU_BACKHAUL_STEER_REQUEST:
			return CMDU_BACKHAUL_STEER_RESPONSE;
		case CMDU_CHANNEL_SCAN_REQUEST:
			return CMDU_CHANNEL_SCAN_REPORT;
		case CMDU_CAC_REQUEST:
			return CMDU_TYPE_NONE;
		case CMDU_CAC_TERMINATION:
			return CMDU_TYPE_NONE;
		case CMDU_CLIENT_DISASSOCIATION_STATS:
			return CMDU_TYPE_NONE;
		case CMDU_ERROR_RESPONSE:
			return CMDU_TYPE_NONE;
		case CMDU_ASSOCIATION_STATUS_NOTIFICATION:
			return CMDU_TYPE_NONE;
		case CMDU_BACKHAUL_STA_CAPABILITY_QUERY:
			return CMDU_BACKHAUL_STA_CAPABILITY_REPORT;
		case CMDU_FAILED_CONNECTION:
			return CMDU_TYPE_NONE;
		default:
			break;
	}

	return CMDU_TYPE_NONE;
}

static uint16_t cntlr_cmdu_expect_response(struct controller *c, uint16_t req_type)
{
	uint16_t resp_type = c_cmdu_expect_response(req_type);

	if (resp_type == CMDU_TYPE_NONE)
		return CMDU_TYPE_NONE;

	if (map_cmdu_mask_isset(c->cmdu_mask, resp_type))
		return resp_type;
	else
		return CMDU_TYPE_NONE;
}

void send_cmdu_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *tmp;
	uint16_t *mid;
	char *str;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message received is NULL\n", __func__);
		return;
	}

	mid = (uint16_t *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "mid", &tmp)) {
		*mid = json_object_get_int(tmp);
		fprintf(stdout, "%s:%d cntlr map-mid:%d\n", __func__, __LINE__, *mid); // typo ||
	}

	json_object_put(jobj);
}

int send_cmdu_ubus(struct controller *a, struct cmdu_buff *cmdu)
{
	struct blob_buf b = { 0 };
	char dst_addr[18] = { 0 };
	uint16_t msgid = 0;
	int ret = 0;
	uint32_t id;

	trace("|%s:%d| Entry\n", __func__, __LINE__);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	blobmsg_add_u32(&b, "type", cmdu_get_type(cmdu));

	hwaddr_ntoa(cmdu->origin, dst_addr);
	blobmsg_add_string(&b, "dst", dst_addr);

	blobmsg_add_u32(&b, "mid", (uint32_t)cmdu_get_mid(cmdu));

	trace("|%s:%d| cmdu:0x%04x| dst:%s| mid:%u| fid:%u\n", __func__, __LINE__,
			cmdu_get_type(cmdu), dst_addr, cmdu_get_mid(cmdu),
			cmdu_get_fid(cmdu));

	if (cmdu->datalen) {
		char *tlv_str = NULL;
		uint16_t len = 0;

		len = (cmdu->datalen * 2) + 1;
		tlv_str = (char *)calloc(len, sizeof(char));
		if (!tlv_str)
			goto out;

		btostr(cmdu->data, cmdu->datalen, tlv_str);
		tlv_str[len-1] = '\0';
		blobmsg_add_string(&b, "data", tlv_str);
		trace("|%s:%d|data:%s|\n", __func__, __LINE__, tlv_str);
		free(tlv_str);
	}

	if (ubus_lookup_id(a->ubus_ctx, "ieee1905", &id)) {
		trace("[%s:%d] not present ieee1905", __func__, __LINE__);
		goto out;
	}

	ret = ubus_invoke(a->ubus_ctx, id, "cmdu",
				b.head, send_cmdu_cb,
				&msgid,
				1000);
	if (ret) {
		trace("[%s:%d] ubus call failed for |ieee1905 send|",
					__func__, __LINE__);
		goto out;
	}

	trace("|%s:%d| msgid = %d\n", __func__, __LINE__, msgid);
	blob_buf_free(&b);
	return msgid;

out:
	blob_buf_free(&b);
	return -1;
}

uint16_t send_cmdu(struct controller *a, struct cmdu_buff *cmdu)
{
	uint16_t resp_type;
	int ret;
	void *cookie = NULL;
	const int resend_num = a->cfg.resend_num;
	uint16_t msgid, old_mid;

	if (hwaddr_is_ucast(cmdu->origin)) {
		resp_type = cntlr_cmdu_expect_response(a, cmdu_get_type(cmdu));
		if (resp_type != CMDU_TYPE_NONE)
			cookie = cmdu_clone(cmdu);
	}

	ret = send_cmdu_ubus(a, cmdu);
	if (ret < 0) {
		err("fail to send cmdu %04x over ubus\n", cmdu_get_type(cmdu));
		goto error;
	}

	msgid = ret;

	old_mid = cmdu_get_mid(cmdu);
	if (old_mid == 0)
		cmdu_set_mid(cmdu, msgid);
	else if (old_mid != msgid)
		warn("msgid differs %d %d for cmdu %04x\n", old_mid, msgid,
		     cmdu_get_type(cmdu));

	if (cookie) {
		ret = cmdu_ackq_enqueue(&a->cmdu_ack_q, resp_type,
					cmdu_get_mid(cmdu), cmdu->origin,
					CMDU_DEFAULT_TIMEOUT, resend_num,
					cookie);
		if (ret < 0) {
			err("cmdu_ackq enqueue failed\n");
			goto error;
		}
	}

	return msgid;

error:
	cmdu_free((struct cmdu_buff *) cookie);
	return 0xffff;
}

struct node *cntlr_add_agent(struct controller *c, uint8_t *hwaddr)
{
	struct node *n;
	char mac_str[18] = {0};

	if (!hwaddr_ntoa(hwaddr, mac_str))
		return NULL;

	n = alloc_node_init(c, hwaddr);
	if (!n) {
		warn("Failed to add node in topology!");
		return NULL;
	}

	cntlr_config_add_agent(&c->cfg, mac_str);

	cntlr_resync_config(c, true);

#ifdef CONTROLLER_SYNC_DYNAMIC_CNTLR_CONFIG
	if (!hwaddr_equal(c->almac, hwaddr))
		cntlr_sync_dyn_controller_config(c, hwaddr);
#endif

	return n;
}

int handle_ap_autoconfig_search(void *cntlr, struct cmdu_buff *rx_cmdu)
{
	trace("%s: --->\n", __func__);

	struct controller *c = (struct controller *) cntlr;
	struct tlv_autoconfig_band *freq;
	struct cmdu_buff *cmdu;
	uint8_t aladdr_origin[6] = {0};
	struct tlv *tv[6][16] = {0};
	int ret = 0, i;
	char freq_band[8] = {0};

	if (!validate_ap_autoconfig_search(rx_cmdu, tv)) {
		dbg("cmdu validation: [AP_AUTOCONFIG_SEARCH] failed\n");
		return -1;
	}

	if (!tv[3][0]) {
		dbg("|%s:%d| Discard ap-autoconfig search due to missing "\
		    "support service TLV\n", __func__, __LINE__);
		return -1;
	}


	memcpy(aladdr_origin, tv[0][0]->data, 6);
	if (hwaddr_is_zero(aladdr_origin)) {
		trace("%s: Discard ap-autoconfig search from aladdr = 0!\n",
			__func__);

		return -1;
	}

	if (tv[1][0]->data[0] != IEEE80211_ROLE_REGISTRAR) {
		trace("%s: Discard ap-autoconfig search for role != registrar\n",
			__func__);
		return -1;
	}

	freq = (struct tlv_autoconfig_band *)tv[2][0]->data;
	if (freq->band > IEEE80211_FREQUENCY_BAND_60_GHZ) {
		trace("%s: Discard ap-autoconfig search for invalid WiFi band\n",
			__func__);
		return -1;
	}

	if (freq->band == IEEE80211_FREQUENCY_BAND_2_4_GHZ) {
		if (!c->cfg.has_registrar_2g)
			return -1;
	} else if (freq->band == IEEE80211_FREQUENCY_BAND_5_GHZ) {
		if (!c->cfg.has_registrar_5g)
			return -1;
	} else
		return -1;

	snprintf(freq_band, sizeof(freq_band), "%s",
			(freq->band == IEEE80211_FREQUENCY_BAND_2_4_GHZ) ? "2" : "5");

	/* if supports agent, add to config */
	for (i = 1; i <= tv[3][0]->data[0]; i++) {
		if (tv[3][0]->data[i] ==
				SUPPORTED_SERVICE_MULTIAP_AGENT) {
			if (c->state == CNTLR_INIT) {
				dbg("|%s:%d| Discard ap-autoconfig search from"\
						" agent during INIT phase\n",
						__func__, __LINE__);
				return -1;
			}
			cntlr_add_agent(c, aladdr_origin);
		} else if (tv[3][0]->data[i] ==
				SUPPORTED_SERVICE_MULTIAP_CONTROLLER) {
			if (!memcmp(aladdr_origin, c->almac, 6)) {
				trace("%s: Discard ap-autoconfig search"\
						" from self\n", __func__);
				return -1;
			}

			if (c->state == CNTLR_INIT) {
				uint32_t uci_obj;
				int res;

				trace("Disable and exit\n");
				set_value_by_string("mapcontroller",
						"controller",
						"enabled", "0",
						UCI_TYPE_STRING);

				res = ubus_lookup_id(c->ubus_ctx, "uci", &uci_obj);
				if (!res) {
					struct blob_buf bb = {0};

					/* trigger commit in order to reload mapcontroller
					 * and not cause crash reports by procd
					 */

					blob_buf_init(&bb, 0);
					blobmsg_add_string(&bb, "config",
							"mapcontroller");
					res = ubus_invoke(c->ubus_ctx, uci_obj,
							"commit", bb.head,
							NULL, NULL,
							2 * 1000);
					if (res)
						err("Failed to get 'commit' (ret = %d), exit anyway\n",
								res);
					blob_buf_free(&bb);
				}

				exit(0); /* TODO: exiting! */
			} else {
				char data[128] = {0};

				/* cppcheck-suppress wrongPrintfScanfArgNum */
				snprintf(data, sizeof(data), "{\"type\":\"error\", \"reason\":\"multiple controllers\", \"data\": {\"remote_almac\":\""MACFMT"\"}}", MAC2STR(rx_cmdu->origin));
				cntlr_notify_event(c, "map.controller", data);
			}
		} else {
			dbg("|%s:%d| Invalid Supported Service value, return\n",
					__func__, __LINE__);
			return -1;
		}

		break;
	}

	trace("%s: sending autoconfig response for band = %d\n",
			__func__, freq->band);
	cmdu = cntlr_gen_ap_autoconfig_response(cntlr, aladdr_origin,
			freq->band,
			cmdu_get_mid(rx_cmdu));
	if (!cmdu)
		return -1;

	ret = send_cmdu(c, cmdu);

	cmdu_free(cmdu);

	UNUSED(ret);

	return 0;
}

/* disable and quit on controller response */
int handle_ap_autoconfig_response(void *cntlr, struct cmdu_buff *rx_cmdu)
{
	trace("%s: --->\n", __func__);
	struct controller *c = (struct controller *) cntlr;
	bool has_cntlr = false;
	struct tlv *tv[4][16] = {0};

	if (!validate_ap_autoconfig_response(rx_cmdu, tv)) {
		dbg("cmdu validation: [AP_AUTOCONFIG_RESPONSE] failed\n");
		return -1;
	}

	if (!tv[1][0]) {
		dbg("|%s:%d| discard AP-Autoconfig response due to missing"\
		    "supported service tlv\n", __func__, __LINE__);
		return -1;
	}

	/* find if supported services containts controller */
	if (tv[1][0]->data[0] > 0) {
		int i;

		for (i = 0; i < tv[1][0]->data[0]; i++) {
			if (tv[1][0]->data[(i+1)] ==
					SUPPORTED_SERVICE_MULTIAP_CONTROLLER) {
				has_cntlr = true;
				break;
			}
		}
	}
	/* if does not support controller - return */
	if (!has_cntlr) {
		dbg("autoconfig response does not support controller!\n");
		return -1;
	}

	/* discard self response */
	if (!memcmp(rx_cmdu->origin, c->almac, 6))
		return 0;

	trace("Received AP-Autoconfig Response which was not from self\n");

	if (c->state == CNTLR_INIT) {
		uint32_t uci_obj;
		int res;

		trace("Disable and exit\n");
		set_value_by_string("mapcontroller", "controller", "enabled", "0",
				UCI_TYPE_STRING);
		res = ubus_lookup_id(c->ubus_ctx, "uci", &uci_obj);
		if (!res) {
			struct blob_buf bb = {0};

			/* trigger commit in order to reload mapcontroller
			 * and not cause crash reports by procd
			 */

			blob_buf_init(&bb, 0);
			blobmsg_add_string(&bb, "config",
					"mapcontroller");
			res = ubus_invoke(c->ubus_ctx, uci_obj,
					"commit", bb.head,
					NULL, NULL,
					2 * 1000);
			if (res)
				err("Failed to get 'commit' (ret = %d), "\
					"exit anyway\n", res);
			blob_buf_free(&bb);
		}

		exit(0);
	} else {
		char data[128] = {0};

		/* cppcheck-suppress wrongPrintfScanfArgNum */
		snprintf(data, sizeof(data), "{\"type\":\"error\", \"reason\":\"multiple controllers\", \"data\": {\"remote_almac\":\""MACFMT"\"}}", MAC2STR(rx_cmdu->origin));
		cntlr_notify_event(c, "map.controller", data);
	}

	return 0;
}

int handle_ap_autoconfig_wsc(void *cntlr, struct cmdu_buff *rx_cmdu)
{
	trace("%s: --->\n", __func__);

	struct controller *c = (struct controller *) cntlr;
	struct node *n;
	struct cmdu_buff *cmdu;
	struct tlv *tv[4][16] = {0};
	bool agent_found;
	struct tlv_ap_radio_basic_cap *ap_caps;
	uint8_t wildcard[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	struct node_policy *np;

	if (!validate_ap_autoconfig_wsc(rx_cmdu, tv)) {
		dbg("cmdu validation: [AP_AUTOCONFIG_WSC] failed\n");
		return -1;
	}

	list_for_each_entry(n, &c->nodelist, list) {
		if (!memcmp(rx_cmdu->origin, n->alid, 6)) {
			agent_found = true;
			break;
		}
	}

	if (!agent_found)
		return -1;

	ap_caps = (struct tlv_ap_radio_basic_cap *) tv[0][0]->data;

	trace("%s: prepare autoconfig wsc response\n", __func__);
	cmdu = cntlr_gen_ap_autoconfig_wsc(cntlr, n->np, rx_cmdu,
			ap_caps->radio, tv[3][0], cmdu_get_mid(rx_cmdu));
	if (!cmdu)
		return -1;

	send_cmdu(c, cmdu);

	cmdu_free(cmdu);

	np = agent_find_policy(c, rx_cmdu->origin);
	if (!np)
		return -1;

	trace("%s: sending policy config request\n", __func__);
	cmdu = cntlr_gen_policy_config_req(cntlr, rx_cmdu->origin, np, 1,
			ap_caps->radio, 1, wildcard);
	if (!cmdu)
		return -1;

	send_cmdu(c, cmdu);
	cmdu_free(cmdu);
	return 0;
}


int handle_1905_ack(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);

	return 0;
}

int handle_ap_caps_report(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);
	trace("parsing AP capabilities of |" MACFMT "|\n", MAC2STR(cmdu->origin));
	struct node *n;
	int index = 0;
	int i = 0;
	int offset = 0;
	struct controller *c = (struct controller *) cntlr;
	struct tlv *tv[9][16];
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_AP_CAPABILITY,
			.present = TLV_PRESENT_ONE
		},
		[1] = {
			.type = MAP_TLV_AP_RADIO_BASIC_CAPABILITIES,
			.present = TLV_PRESENT_MORE
		},
		[2] = {
			.type = MAP_TLV_AP_HT_CAPABILITIES,
			.present = TLV_PRESENT_MORE
		},
		[3] = {
			.type = MAP_TLV_AP_VHT_CAPABILITIES,
			.present = TLV_PRESENT_MORE
		},
		[4] = {
			.type = MAP_TLV_AP_HE_CAPABILITIES,
			.present = TLV_PRESENT_MORE
		},
		[5] = {
			.type = MAP_TLV_CHANNEL_SCAN_CAPABILITY,
			.present = TLV_PRESENT_ONE
		},
		[6] = {
			.type = MAP_TLV_CAC_CAPABILITY,
			.present = TLV_PRESENT_ONE
		},
		[7] = {
			.type = MAP_TLV_PROFILE2_AP_CAPABILITY,
			.present = TLV_PRESENT_ONE
		},
		[8] = {
			.type = MAP_TLV_METRIC_COLLECTION_INTERVAL,
			.present = TLV_PRESENT_ONE
		}
	};

	cmdu_parse_tlvs(cmdu, tv, d_policy, 9);
	if (!tv[1][0])
		return -1;

	n = find_node_by_mac(c, cmdu->origin);
	if (!n)
		return -1;

	/* AP Capability TLV */
	if (tv[0][0]) {
		struct tlv_ap_cap *p = (struct tlv_ap_cap *)tv[0][0]->data;

		dbg("%s %d AP capability is 0x%02x\n", __func__, __LINE__, p->cap);
		n->ap_cap = p->cap;
	}

	index = 0;
	/* Parse AP Radio Basic Capabilities TLV */
	while (tv[1][index] && (index < 16)) {
		struct netif_radio *r;
		uint8_t *tv_data = (uint8_t *)tv[1][index++]->data;
		struct tlv_ap_radio_basic_cap *p =
			(struct tlv_ap_radio_basic_cap *)tv_data;

		r = find_radio_by_node(c, n, p->radio);
		if (!r)
			continue;

		offset = sizeof(*p);

		(void) offset;
		(void) i;
#if 0
		/* TODO change this - this decribe exclooded channels */
		r->num_opclass = p->num_opclass;
		for (i = 0; i < p->num_opclass; i++) {
			struct ap_radio_basic_cap_opclass *op =
				(struct ap_radio_basic_cap_opclass *)&tv_data[offset];
			r->opclass[i] = op->classid;
			offset += sizeof(*op) + op->num_nonop_channel;
		}

		/* TODO: very crude, please fix */
		if (r->num_opclass && r->opclass[0] > 81)
			r->band = BAND_5;
		else if (r->num_opclass && r->opclass[0] == 81)
			r->band = BAND_2;
#endif

		dbg("%s %d radio "MACFMT" is on band %d\n", __func__, __LINE__, MAC2STR(r->macaddr), r->band);
	}

	return 0;
}

int handle_channel_pref_report(void *cntlr, struct cmdu_buff *cmdu)
{
	int idx, offset = 0;
	int i, j;
	struct tlv_policy pref_policy[] = {
		[0] = { .type = MAP_TLV_CHANNEL_PREFERENCE, .present = TLV_PRESENT_MORE },
		[1] = { .type = MAP_TLV_RADIO_OPERATION_RESTRICTION, .present = TLV_PRESENT_MORE },
		[2] = { .type = MAP_TLV_CAC_COMPLETION_REPORT, .present = TLV_PRESENT_MORE },
		[3] = { .type = MAP_TLV_CAC_STATUS_REPORT, .present = TLV_PRESENT_ONE },
	};
	struct tlv *tv[ARRAY_SIZE(pref_policy)][16] = { 0 };
	struct netif_radio *radio;
	struct node *node;

	trace("%s: --->\n", __func__);
	node = find_node_by_mac(cntlr, cmdu->origin);
	if (!node)
		return -1;

	cmdu_parse_tlvs(cmdu, tv, pref_policy, 4);

	idx = 0;
	while (tv[0][idx]) {
		struct tlv *t = (struct tlv *)tv[0][idx++];
		uint8_t mac[6] = { 0 };
		int num_opclass;

		offset = 0;
		memcpy(mac, &t->data[offset], 6);
		offset += 6;
		num_opclass = t->data[offset++];

		radio = find_radio_by_node(cntlr, node, mac);
		if (!radio)
			continue;

		/* TODO skip this if unsolicited report received */
		cntlr_radio_opclass_reset(radio);

		for (i = 0; i < num_opclass; i++) {
			uint8_t opclass;
			uint8_t num_channel;
			uint8_t preference;
			uint8_t channel;

			opclass = t->data[offset++];
			num_channel = t->data[offset++];
			preference = t->data[offset + num_channel];

			for (j = 0; j < num_channel; j++) {
				channel = t->data[offset++];
				if (cntlr_radio_opclass_add(radio, opclass, channel, preference))
					warn("opclass_add %u %u %u failed\n", opclass, channel, preference);
			}

			/* last preference - simple move offset */
			offset++;

		}
		cntlr_radio_opclass_dump(radio);
	}

	return 0;
}

int handle_channel_sel_response(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);
	return 0;
}

int handle_oper_channel_report(void *cntlr, struct cmdu_buff *cmdu)
{
	int idx = 0;
	struct tlv *tv[1][16];
	struct tlv_policy op_policy[] = {
		[0] = { .type = MAP_TLV_OPERATING_CHANNEL_REPORT, .present = TLV_PRESENT_MORE }
	};
	struct controller *c = (struct controller *) cntlr;
	struct node *n;

	trace("%s: --->\n", __func__);

	n = find_node_by_mac(c, cmdu->origin);
	if (!n)
		return -1;

	cmdu_parse_tlvs(cmdu, tv, op_policy, 1);

	while (tv[0][idx]) {
		int i;
		int offset = 0;
		uint8_t mac[6] = {0};
		uint8_t *p = (uint8_t *)tv[0][idx++]->data;
		struct netif_radio *r;
		uint8_t channel;
		uint8_t opclass;
		uint8_t txpower;
		int num_opclass;

		memcpy(mac, &p[offset], 6);
		offset += 6;

		r = find_radio_by_node(c, n, mac);
		if (!r) {
			r = cntlr_node_add_radio(c, n, mac);
			if (!r)
				continue;
		}

		/* Reset current settings */
		cntlr_radio_cur_opclass_reset(r);

		num_opclass = p[offset++];
		txpower = p[offset + 2 * num_opclass];
		for (i = 0; i < num_opclass; i++) {
			opclass = p[offset++];
			channel = p[offset++];
			if (WARN_ON(cntlr_radio_cur_opclass_add(r, opclass, channel, txpower)))
				continue;
		}
		offset++;

		/* Upate band based on current opclass */
		cntlr_radio_update_band(r);
		cntlr_radio_cur_opclass_dump(r);
	}

	return 0;
}

int handle_sta_caps_report(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);
//	trace("parsing AP capabilities of |%s:" MACFMT "|\n",
//			cmdu->intf_name, MAC2STR(cmdu->origin));
	return 0;
}

int handle_ap_metrics_response(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);
	int idx, index;
	struct controller *c = (struct controller *) cntlr;
	//struct node *n;
	//struct netif_radio *r;
	struct netif_iface *ifc;
	struct bss_metrics *a;
	//struct link_metrics *link;
	struct tlv *tv[6][16] = { 0 };
	struct tlv_policy metric_policy[] = {
		[0] = { .type = MAP_TLV_AP_METRICS, .present = TLV_PRESENT_MORE },
		[1] = { .type = MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS, .present = TLV_PRESENT_MORE },
		[2] = { .type = MAP_TLV_ASSOCIATED_STA_LINK_METRICS, .present = TLV_PRESENT_MORE },
		[3] = { .type = MAP_TLV_AP_EXTENDED_METRICS, .present = TLV_PRESENT_MORE },
		[4] = { .type = MAP_TLV_RADIO_METRICS, .present = TLV_PRESENT_MORE },
		[5] = { .type = MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS, .present = TLV_PRESENT_MORE }
	};
	struct tlv_ap_metrics *p;

	if (!cmdu) {
		trace("error: NO CMDU\n");
		return -1;
	}

	trace("Storing AP metrics of |" MACFMT "|\n", MAC2STR(cmdu->origin));

	/* Parsing AP Metrics TLV */
	cmdu_parse_tlvs(cmdu, tv, metric_policy, 6);

	/* Storing AP Metrics TLV */
	idx = 0;
	while (tv[0][idx]) {
		p = (struct tlv_ap_metrics *) tv[0][idx++]->data;
		if (!p)
			continue;

		ifc = cntlr_iterate_fbss(c, p->bssid);
		if (!ifc)
			continue;

		a = ifc->metrics;
		index = 0;
		memcpy(a->bssid, p->bssid, 6);
		a->ch_util = p->channel_utilization;
		a->num_sta = p->num_station;
		a->esp_ac = p->esp_ac;
		if (p->esp_ac & ESP_AC_BE) {
			memcpy(a->esp + index, p->esp + index, 3);
		}
		if (p->esp_ac & ESP_AC_BK) {
			memcpy(a->esp + index, p->esp + index, 3);
			index += 3;
		}
		if (p->esp_ac & ESP_AC_VO) {
			memcpy(a->esp + index, p->esp + index, 3);
			index += 3;
		}
		if (p->esp_ac & ESP_AC_VI) {
			memcpy(a->esp + index, p->esp + index, 3);
		}
	}

		/* trying to print out what is stored */
		// size_t out_len;
		// index = 0;
		// list_for_each_entry(n, &c->nodelist, list) {
		// list_for_each_entry(r, &n->radiolist, list) {
		// list_for_each_entry(ifc, &r->iflist, list) {

		// a = ifc->metrics;
		// unsigned char est_str[16];
		// trace("STORED:\n");
		// trace("\tbssid: " MACFMT "\n", MAC2STR(a->bssid));
		// trace("\tchannel_utilization: %d\n", a->ch_util);
		// trace("\tnum_station: %d\n", BUF_GET_BE16(a->num_sta));
		// trace("\tesp_ac: %d\n", a->esp_ac);
		// if (a->esp_ac & ESP_AC_BE) {
		// 	out_len = sizeof(est_str);
		// 	memset(est_str, 0, sizeof(est_str));
		// 	base64_encode(a->esp_be, 3, est_str, &out_len);
		// 	trace("\tservice_param_info_be:%s\n", (char *)est_str);
		// }
		// if (a->esp_ac & ESP_AC_BK) {
		// 	out_len = sizeof(est_str);
		// 	memset(est_str, 0, sizeof(est_str));
		// 	base64_encode(a->esp + index, 3, est_str, &out_len);
		// 	trace("\tservice_param_info_bk:%s\n", (char *)est_str);
		// 	index += 3;
		// }
		// if (a->esp_ac & ESP_AC_VO) {
		// 	out_len = sizeof(est_str);
		// 	memset(est_str, 0, sizeof(est_str));
		// 	base64_encode(a->esp + index, 3, est_str, &out_len);
		// 	trace("\tservice_param_info_vo:%s\n", (char *)est_str);
		// 	index += 3;
		// }
		// if (a->esp_ac & ESP_AC_VI) {
		// 	out_len = sizeof(est_str);
		// 	memset(est_str, 0, sizeof(est_str));
		// 	base64_encode(a->esp + index, 3, est_str, &out_len);
		// 	trace("\tservice_param_info_vi:%s\n", (char *)est_str);
		// }
		// }}}

	return 0;
}

static int cntlr_request_usta_metrics(struct controller *c,
		struct node *n, struct sta *s)
{
	struct cmdu_buff *usta_cmdu = NULL;
	struct unassoc_sta_metric metrics[1] = {};
	struct netif_radio *r;
	struct netif_iface *fh;
	int j;

	trace("%s: --->\n", __func__);

	if (!n || !s)
		return -1;

	trace("%s %d ap_cap = 0x%02x\n", __func__, __LINE__, n->ap_cap);

	if (n->ap_cap & UNASSOC_STA_REPORTING_OFFCHAN)
		/* TODO: off channel measurements */
		dbg("%s: Offchannel Unassoc STA metric not supported");

	if (n->ap_cap & UNASSOC_STA_REPORTING_ONCHAN) {

		r = find_radio_by_ssid(c, n, s->fh->ssid);
		fh = find_interface_by_ssid(c, n, s->fh->ssid);

		if (!r || !fh)
			return -1;

		for (j = 0; j < r->cur_opclass.opclass_entry_num; j++) {
			metrics[0].channel = fh->channel;
			metrics[0].num_sta = 1;
			memcpy(metrics[0].sta[0].macaddr, s->macaddr, 6);

			usta_cmdu = cntlr_gen_unassoc_sta_metric_query(c,
					n->alid, r->cur_opclass.opclass_entry[j].opclass, 1, metrics);

			if (usta_cmdu) {
				send_cmdu(c, usta_cmdu);
				cmdu_free(usta_cmdu);
			}
		}

	} else {
		dbg("%s: Unassoc STA metric not supported by " MACFMT "\n",
		    __func__, MAC2STR(n->alid));
	}

	return 0;
}

static void cntlr_request_bcn_metrics_bsta(struct controller *c, struct sta *s)
{
	struct cmdu_buff *bcn_cmdu;
	uint8_t wildcard[6] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

	dbg("%s: --->\n", __func__);

	bcn_cmdu = cntlr_gen_beacon_metrics_query(c,
			s->fh->agent->alid, s->macaddr, 0, 0,
			wildcard, 0, s->fh->ssid, 0, NULL, 0, NULL);

	if (bcn_cmdu) {
		send_cmdu(c, bcn_cmdu);
		cmdu_free(bcn_cmdu);
	}
}

static uint8_t cntlr_get_opclass_ht20(struct netif_radio *nr, uint8_t channel)
{
    int i, j;
	struct opclass_entry *entry;

	for (i = 0; i < nr->opclass.opclass_entry_num; i++) {
		entry = &nr->opclass.opclass_entry[i];

		if (entry->bw != 20)
			continue;

		for (j = 0; j < entry->channels_num; j++) {
			if (entry->channels[j].channel == channel)
				return entry->opclass;
		}
	}

	return 0; /* Not found */
}

static int cntlr_request_bcn_metrics_sta(struct controller *c, struct sta *s)
{
	struct cmdu_buff *bcn_cmdu;
	uint8_t wildcard[6] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
	uint8_t opclass = 0;
	uint8_t channel = 0;
	struct sta_channel_report *reports = NULL;
	struct netif_radio *nr;
	struct steering *sp;
	uint8_t num_report = 0;
	struct {
		uint8_t	opclass;
		uint8_t	channel;
	} op_ch[16];
	uint8_t opc;
	int i;

	trace("%s: --->\n", __func__);

	nr = find_radio_by_bssid(c, s->bssid);
	if (!nr)
		return -1;

	sp = &c->steer_params;

	dbg("|%s:%d| channels_num = %d\n", __func__, __LINE__, sp->channels_num);

	if (sp->channels_num < 1 || sp->channels_num > 16)
		return -1;

	if (sp->channels_num == 1) {
		/* won't use channel report */
		opclass = cntlr_get_opclass_ht20(nr, sp->channels[0]); /* /20 */
		if (!opclass)
			return -1;
		channel = sp->channels[0];
		num_report = 0;
	} else {
		opclass = 0;
		channel = 255; /* use channel report */
		for (i = 0; i < sp->channels_num; i++) {
			opc = cntlr_get_opclass_ht20(nr, sp->channels[i]);
			if (!opc)
				continue;
			op_ch[num_report].opclass = opc;
			op_ch[num_report].channel = sp->channels[i];
			num_report++;
		}

		if (!num_report)
			return -1;

		reports = calloc(num_report, sizeof(struct sta_channel_report));
		if (!reports)
			return -ENOMEM;

		for (i = 0; i < num_report; i++) {
			reports[i].opclass = op_ch[i].opclass;
			reports[i].num_channel = 1;
			reports[i].channel[0] = op_ch[i].channel;
		}
	}

	dbg("|%s:%d| alid " MACFMT " s->macaddr " MACFMT " num_report = %d\n",
	    __func__, __LINE__,
	    MAC2STR(s->fh->agent->alid),
	    MAC2STR(s->macaddr),
	    num_report);

	bcn_cmdu = cntlr_gen_beacon_metrics_query(c,
			s->fh->agent->alid, s->macaddr,
			opclass, channel,
			wildcard, 0, s->fh->ssid,
			num_report, reports, 0, NULL);

	if (bcn_cmdu) {
		send_cmdu(c, bcn_cmdu);
		cmdu_free(bcn_cmdu);
	}

	if (reports)
		free(reports);

	/* returns expected number of requests in agents */
	return (num_report ? num_report : 1);
}

int handle_sta_link_metrics_response(void *cntlr, struct cmdu_buff *cmdu)
{
	int i;
	int offset = 0;
	struct controller *c = (struct controller *) cntlr;
	struct tlv *tv[3][16] = { 0 };
	struct node *n;
	struct tlv_policy sta_policy[] = {
		[0] = { .type = MAP_TLV_ASSOCIATED_STA_LINK_METRICS, .present = TLV_PRESENT_MORE },
		[1] = { .type = MAP_TLV_ERROR_CODE, .present = TLV_PRESENT_MORE },
		[2] = { .type = MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS, .present = TLV_PRESENT_MORE }
	};

	trace("%s: --->\n", __func__);
	dbg("parsing sta link metric response |" MACFMT "\n",
			MAC2STR(cmdu->origin));
	cmdu_parse_tlvs(cmdu, tv, sta_policy, 3);

	n = find_node_by_mac(c, cmdu->origin);
	if (!n)
		return -1;

	if (tv[0][0]) {
		struct sta *s;
		uint8_t *tv_data = (uint8_t *)tv[0][0]->data;
		struct tlv_assoc_sta_link_metrics *p =
			(struct tlv_assoc_sta_link_metrics *)tv_data;
		struct netif_radio *r;
		struct radio_policy *rp;

		s = cntlr_find_sta(c, p->macaddr);
		if (!s)
			return -1;

		offset = sizeof(*p);
		for (i = 0; i < p->num_bss; i++) {
			struct assoc_sta_link_metrics_bss *b =
				(struct assoc_sta_link_metrics_bss *)&tv_data[offset];

			memcpy(s->bssid, b->bssid, 6);
			s->dl_thput = BUF_GET_BE32(b->dl_thput);
			s->ul_thput = BUF_GET_BE32(b->ul_thput);
			s->time_delta = BUF_GET_BE32(b->time_delta);
			s->ul_rcpi = b->ul_rcpi;

			offset += sizeof(*b);
		}

		r = find_radio_by_bssid(c, s->bssid);
		if (!r)
			return -1;

		rp = agent_find_radio_policy(c, r->macaddr);
		if (!rp)
			return -1;

		if (s->type == IEEE1905 && s->ul_rcpi < rp->rcpi_threshold) {

			if (!c->cfg.enable_bsta_steer) {
				dbg("|%s:%d| rcpi below %d, but will not query for any " \
				    "metrics as 'enable_bsta_steer' is not set!\n",
				    __func__, __LINE__, rp->rcpi_threshold);
				return 0;
			}

			/* Get bcn metrics for bsta */
			if (c->cfg.use_bcn_metrics) {
				dbg("|%s:%d| request bcn metrics from bsta\n",
				    __func__, __LINE__);
				free_bcn_metrics(c, s);
				cntlr_request_bcn_metrics_bsta(c, s);
				/* TODO: set timeout here (?) */
			}

		} else if (s->type == NON_IEEE1905 && s->ul_rcpi < rp->rcpi_threshold) {

			if (!c->cfg.enable_sta_steer) {
				dbg("|%s:%d| rcpi below %d, but will not query for any " \
				    "metrics as 'enable_sta_steer' is not set!\n",
				    __func__, __LINE__, rp->rcpi_threshold);
				return 0;
			}

			/* Get bcn metrics */
			if (c->cfg.use_bcn_metrics) {
				int num_req = 0;

				/* FIXME: consider dynamic timer value */
				if (timestamp_expired(&s->last_bcn_metrics_query, 20 * 1000)) {
					dbg("|%s:%d| requesting bcn metrics from sta\n",
					    __func__, __LINE__);
					free_bcn_metrics(c, s);
					num_req = cntlr_request_bcn_metrics_sta(c, s);
					if (num_req > 0) {
						timestamp_update(&s->last_bcn_metrics_query);
						uloop_timeout_set(&s->bcn_metrics_timer, num_req * 3 * 1000);
					}
				} else {
					dbg("|%s:%d| skip requesting bcn metrics - up to date\n",
					    __func__, __LINE__);
				}
			}

			/* Get usta metrics for each agent in mesh */
			if (c->cfg.use_usta_metrics) {
				struct node *no = NULL;

				dbg("|%s:%d| request usta metrics from sta\n",
				      __func__, __LINE__);
				free_usta_metrics(c, s);
				list_for_each_entry(no, &c->nodelist, list) {
					cntlr_request_usta_metrics(c, no, s);
				}
			}
		}
	}

	return 0;
}

int cntlr_send_1905_acknowledge(void *cntlr,
	struct cmdu_buff *rx_cmdu,
	struct sta_error_response *sta_resp, uint32_t sta_count)
{
	struct cmdu_buff *response;
	struct controller *c = (struct controller *) cntlr;

	trace("%s: --->\n", __func__);

	response = cntlr_gen_cmdu_1905_ack(c, rx_cmdu, sta_resp, sta_count);
	if (!response)
		return -1;

	send_cmdu(c, response);
	cmdu_free(response);
	return 0;
}

struct una_sta_metrics *cntlr_find_usta_metric(struct controller *c,
		struct sta *s, uint8_t *mac)
{
	struct una_sta_metrics *u = NULL;

	list_for_each_entry(u, &s->unassoclist, list) {
		dbg("%s %d mac "MACFMT" alid " MACFMT"\n", __func__, __LINE__,
				MAC2STR(u->agent->alid), MAC2STR(mac));
		if (!memcmp(u->agent->alid, mac, 6))
			return u;
	}

	return NULL;
}

#define USTA_STEER_UL_RCPI_DELTA 10

/* use unassociated STA measurements to steer */
void cntlr_check_usta_steer(struct controller *c, struct sta *s)
{
	struct una_sta_metrics *best = NULL, *u = NULL;
	struct node *n = s->fh->agent;
	struct netif_iface *best_fh;

	if (!c->cfg.use_usta_metrics) {
		dbg("%s %d Will not use unassociated STA metrics \
		    data to steer\n", __func__, __LINE__);
		return;
	}

	dbg("%s %d for "MACFMT" attached to bssid " MACFMT " node = " \
	    MACFMT "\n", __func__, __LINE__, MAC2STR(s->macaddr),
	    MAC2STR(s->bssid), MAC2STR(n->alid));

	list_for_each_entry(u, &s->unassoclist, list) {
		dbg("%s %d check usta node "MACFMT"\n",
		    __func__, __LINE__, MAC2STR(u->agent->alid));

		if (!best) {
			best = u;
			continue;
		}

		dbg("%s %d best ul_rcpi %u this ul_rcpi %u\n", __func__, __LINE__,
				best->ul_rcpi, u->ul_rcpi);

		if ((best->ul_rcpi - u->ul_rcpi) > USTA_STEER_UL_RCPI_DELTA) {
			dbg("%s %d new best usta node "MACFMT" with ul_rcpi %d\n",
					__func__, __LINE__,
					MAC2STR(u->agent->alid),
					u->ul_rcpi);
			best = u;
		}
	}

	if (!best)
		return;

	/* Get appropriate netif on best node based on current ssid */
	best_fh = find_interface_by_ssid(c, best->agent, s->fh->ssid);

	if (best_fh && !hwaddr_is_zero(best_fh->bssid)
			&& memcmp(best_fh->bssid, s->bssid, 6)) {

		struct cmdu_buff *cmdu;

		if ((s->type == IEEE1905 && !c->cfg.enable_bsta_steer) ||
		    (s->type == NON_IEEE1905 && !c->cfg.enable_sta_steer)) {
			trace("|%s:%d| better bssid found, but will not steer "MACFMT",\
			       because the 'enable_(b)sta_steer' is not set!\n",
			       __func__, __LINE__, MAC2STR(s->macaddr));
			return;
		}

		dbg("%s %d better bssid found! try to steer " MACFMT " \
		    from " MACFMT " to " MACFMT "\n",
		    __func__, __LINE__,
		    MAC2STR(s->macaddr), MAC2STR(s->bssid), MAC2STR(best_fh->bssid));

		if (s->type == IEEE1905)
			cmdu = cntlr_gen_backhaul_steer_request(c, s->agent->alid,
					s->macaddr, best_fh->bssid, 0, 0);
		else
			cmdu = cntlr_gen_client_steer_request(c, s->fh->agent->alid,
					s->bssid, 0,
					1, (uint8_t (*)[6])s->macaddr,
					1, (uint8_t (*)[6])best_fh->bssid,
					1);

		if (cmdu) {
			send_cmdu(c, cmdu);
			cmdu_free(cmdu);
		}
	}
}

int handle_unassoc_sta_link_metrics_response(void *cntlr,
		struct cmdu_buff *cmdu)
{
	struct controller *c = (struct controller *) cntlr;
	struct tlv_policy a_policy[] = {
		[0] = { .type = MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE, .present = TLV_PRESENT_ONE },
	};
	struct tlv *tv[1][16] = {0};
	int i = 0;
	int ret = 0;

	trace("%s: --->\n", __func__);

	if (!cntlr || !cmdu)
		return -1;

	/* If a Multi-AP Controller receives an Unassociated STA
	 * Link Metrics Response message, then it shall respond
	 * within one second with a 1905 Ack message.
	 */
	cntlr_send_1905_acknowledge(cntlr, cmdu, NULL, 0);

	ret = cmdu_parse_tlvs(cmdu, tv, a_policy, 1);

	if (ret || !tv[0][0])
		return -1;


	if (tv[0][0]) {
		int offset = 0;
		uint8_t *tv_data = (uint8_t *)tv[0][0]->data;
		struct tlv_unassoc_sta_link_metrics_resp *resp =
			(struct tlv_unassoc_sta_link_metrics_resp  *)tv_data;

		offset = sizeof(*resp);
		for (i = 0; i < resp->num_sta; i++) {
			struct una_sta_metrics *u;
			struct unassoc_sta_link_metrics_sta *b =
				(struct unassoc_sta_link_metrics_sta *)&tv_data[offset];
			struct sta *s = cntlr_find_sta(c, b->macaddr);

			if (!s) {
				dbg("|%s:%d| unassociated STA "MACFMT" not found!\n",
				__func__, __LINE__, MAC2STR(b->macaddr));
				continue;
			}

			u = cntlr_find_usta_metric(c, s, cmdu->origin); /* alid */
			if (!u) {
				u = calloc(1, sizeof(*u));
				if (!u)
					continue;

				list_add(&u->list, &s->unassoclist);
				u->agent = find_node_by_mac(c, cmdu->origin);
				dbg("%s %d\n", __func__, __LINE__);
			}

			u->channel = b->channel;
			u->time_delta = BUF_GET_BE32(b->time_delta);
			u->ul_rcpi = b->ul_rcpi;

			trace("\t\tu->agent.alid: " MACFMT "\n", MAC2STR(u->agent->alid));
			trace("\t\tu->channel: %d\n", u->channel);
			trace("\t\tu->time_delta: %lu\n",BUF_GET_BE32(u->time_delta));
			trace("\t\tu->ul_rcpi: %d\n", u->ul_rcpi);

			offset += sizeof(*b);

			/* Steer if (b)STA would benefit from it */
			cntlr_check_usta_steer(c, s);
		}
	}

	return ret;
}

struct bcn_metrics *cntlr_find_sta_bcn_metric(struct controller *c,
		struct sta *s, uint8_t *mac)
{
	struct bcn_metrics *b = NULL;

	dbg("%s %d\n", __func__, __LINE__);

	list_for_each_entry(b, &s->bcnlist, list) {
		dbg("%s %d mac "MACFMT" bssid " MACFMT"\n", __func__, __LINE__,
				MAC2STR(b->bssid), MAC2STR(mac));
		if (!memcmp(b->bssid, mac, 6))
			return b;
	}

	return NULL;
}

int handle_beacon_metrics_response(void *cntlr, struct cmdu_buff *cmdu)
{
	struct controller *c = (struct controller *) cntlr;
	struct tlv_policy a_policy[] = {
		[0] = { .type = MAP_TLV_BEACON_METRICS_RESPONSE, .present = TLV_PRESENT_ONE },
	};
	struct tlv_beacon_metrics_resp *resp;
	struct tlv *tv[1][16] = {0};
	uint8_t *ppos;
	struct sta *s;
	struct bcnreq *br;
	bool requested = false;
	int i, ret = 0;

	trace("%s: --->\n", __func__);

	if (!cntlr || !cmdu)
		return -1;

	/* If a Multi-AP Controller receives a Beacon Metrics Response message,
	 * then it shall respond within one second with a 1905 Ack message.
	 */
	cntlr_send_1905_acknowledge(cntlr, cmdu, NULL, 0);

	ret = cmdu_parse_tlvs(cmdu, tv, a_policy, 1);

	if (ret || !tv[0][0])
		return -1;

	resp = (struct tlv_beacon_metrics_resp *) tv[0][0]->data;

	br = cntlr_find_bcnreq(c, resp->sta_macaddr, cmdu->origin);
	if (br || !timestamp_expired(&br->tsp, 10000))
		/* matching request found for this response */
		requested = true;
	else {
		dbg("|%s:%d| no active request for metrics "\
			"of STA "MACFMT" on agent "MACFMT"\n",
			__func__, __LINE__,
			MAC2STR(resp->sta_macaddr),
			MAC2STR(cmdu->origin));
		requested = false;
	}

	s = cntlr_find_sta(c, resp->sta_macaddr);
	if (!s) {
		dbg("|%s:%d| beacon metrics client "MACFMT" not found!\n",
				__func__, __LINE__, MAC2STR(resp->sta_macaddr));
		return -1;
	}

	//s->num_bcn_metrics = resp->num_element;

	ppos = resp->element;
	for (i = 0; i < resp->num_element; i++) {
		struct bcn_meas_element *elem;
		struct bcn_metrics *b;

		elem = (struct bcn_meas_element *) ppos;

		if (elem->tag_number != 0x27) {
			trace("|%s:%d| received beacon report with" \
				"unexpected data\n", __func__, __LINE__);
			return -1;
		}

		if (hwaddr_is_zero(elem->bssid)) {
			/* Move to the next measurement report */
			ppos = ppos + elem->tag_length + 2;
			continue;
		}

		b = cntlr_find_sta_bcn_metric(c, s, elem->bssid);
		if (!b) {
			dbg("%s %d\n", __func__, __LINE__);
			b = calloc(1, sizeof(*b));
			if (!b)
				continue;

			list_add(&b->list, &s->bcnlist);
			memcpy(b->bssid, elem->bssid, 6);
		}

		b->channel = elem->channel;
		b->opclass = elem->op_class;
		b->rcpi = elem->rcpi;
		b->rsni = elem->rsni;
		b->requested = requested;

		/* Move to the next measurement report */
		ppos = ppos + elem->tag_length + 2;
	}

	return ret;
}

static struct steer_attempt *sta_lookup_steer_attempt(struct sta *s,
		uint8_t *src, uint8_t *dst)
{
	int i;
	int size = (s->num_steer_attempts < MAX_LOG_STEER_ATTEMPTS) ?
				s->num_steer_attempts : MAX_LOG_STEER_ATTEMPTS;

	trace("%s: --->\n", __func__);

	/* Target BSSID is empty (error): use most recent unanswered attempt */
	if (!dst) {
		for (i = size - 1; i >= 0 ; i--) {
			if (!s->attempts[i].answered &&
					!memcmp(s->attempts[i].src_bssid, src, 6)) {
				dbg("|%s:%d| empty dst bssid - updated first unused attempt!\n",
				    __func__, __LINE__);
				return &s->attempts[i];
			}
		}
		return NULL;
	}

	dbg("|%s:%d| src="MACFMT", dst="MACFMT"\n",
	    __func__, __LINE__, MAC2STR(src), MAC2STR(dst));

	/* Target BSSID non-empty - lookup attempt based on that */
	for (i = 0; i < size; i++) {
		if (!s->attempts[i].answered &&
				!memcmp(s->attempts[i].src_bssid, src, 6) &&
				!memcmp(s->attempts[i].dst_bssid, dst, 6))
			return &s->attempts[i];
	}

	dbg("|%s:%d| dst "MACFMT" not found on attempt list!\n",
		__func__, __LINE__, MAC2STR(dst));

	/* Target BSSID not found. Check for auto bssid attempt & ignore dst. */
	for (i = 0; i < size; i++) {
		if (!s->attempts[i].answered &&
				!memcmp(s->attempts[i].src_bssid, src, 6) &&
				s->attempts[i].auto_bssid)
			return &s->attempts[i];
	}

	dbg("|%s:%d| auto bssid not found on attempt list!\n",
		__func__, __LINE__);

	return NULL;
}

int handle_sta_steer_btm_report(void *cntlr, struct cmdu_buff *cmdu)
{
	struct controller *c = (struct controller *) cntlr;
	struct tlv_policy a_policy[] = {
		[0] = { .type = MAP_TLV_STEERING_BTM_REPORT, .present = TLV_PRESENT_ONE },
	};
	struct tlv_steer_btm_report *resp;
	struct tlv *tv[1][16] = {0};
	struct sta *s;
	struct steer_attempt *attempt;
	int ret = 0;

	trace("%s: --->\n", __func__);

	if (!cntlr || !cmdu)
		return -1;

	/* If a Multi-AP Controller receives a Client Steering BTM Report message,
	 * then it shall respond within one second with a 1905 Ack message.
	 */
	cntlr_send_1905_acknowledge(cntlr, cmdu, NULL, 0);

	ret = cmdu_parse_tlvs(cmdu, tv, a_policy, 1);

	if (ret || !tv[0][0])
		return -1;

	resp = (struct tlv_steer_btm_report *) tv[0][0]->data;

	s = cntlr_find_sta(c, resp->sta_macaddr);
	if (!s) {
		dbg("|%s:%d| steer btm report: sta "MACFMT" not found!\n",
		    __func__, __LINE__, MAC2STR(resp->sta_macaddr));
		return -1;
	}

	attempt = sta_lookup_steer_attempt(s, resp->bssid,
				resp->status ? NULL : resp->target_bssid[0]);

	if (!attempt)
		dbg("|%s:%d| could not find an attempt matching this btm report\n",
		    __func__, __LINE__);
	else
		attempt->answered = true;

	if (resp->status) {
		s->failed_steer_attempts++;
		/* Update Steering Summary Stats */
		s->stats.btm_failure_cnt++;

		if (attempt)
			/* 'A failed steering attempt will leave this parameter 0' */
			attempt->duration = 0;
	} else {
		s->failed_steer_attempts = 0;
		/* Update Steering Summary Stats */
		s->stats.btm_success_cnt++;

		if (attempt)
			attempt->duration = timestamp_elapsed_sec(&attempt->time);
	}

	/* Number of asynchronous BTM Queries for which a BTM
	 * Request was issued to this Associated Device.
	 */
	s->stats.btm_query_resp_cnt++;

	return ret;
}

int handle_sta_steer_complete(void *cntlr, struct cmdu_buff *cmdu)
{
	/* TODO: implement CMDU_STEERING_COMPLETED */
	trace("%s: --->\n", __func__);
	return 0;
}

static int cntlr_hld_event(struct controller *c, uint8_t proto, uint8_t *data,
			int data_len)
{
	const int len = data_len*2 + 128;
	char *str;
	int idx;

	str = calloc(len, sizeof(char));
	if (!str)
		return -1;

	idx = snprintf(str, len, "{\"protocol\":%d,\"data\":\"", proto);
	btostr(data, data_len, str + idx);
	idx += data_len*2;
	snprintf(str + idx, len - idx, "\"}");

	cntlr_notify_event(c, "map.controller.higher_layer_data", str);

	free(str);

	return 0;
}

int handle_hld_message(void *cntlr, struct cmdu_buff *rx_cmdu)
{
	struct controller *c = (struct controller *) cntlr;
	struct tlv *t;
	uint8_t proto;
	uint8_t *data;
	int data_len;
	int ret;
	uint8_t origin[6] = {0};
	struct node *n = NULL;

	trace("%s: --->\n", __func__);

	t = cmdu_peek_tlv(rx_cmdu, MAP_TLV_HIGHER_LAYER_DATA);
	if (!t) {
		dbg("%s: higher layer data TLV not found\n", __func__);
		return -1;
	}

	data_len = tlv_length(t) - 1;
	proto = t->data[0];
	data = t->data + 1;

	dbg("%s TLV received proto %u data_len %u!!\n", __func__, proto, data_len);

	if (!hwaddr_is_zero(rx_cmdu->origin)) {
		memcpy(origin, rx_cmdu->origin, 6);

		n = find_node_by_mac(c, origin);
		if (!n)
			return -1;
	}

	dbg("Received HLD from " MACFMT "\n", MAC2STR(origin));

	ret = cntlr_hld_event(c, proto, data, data_len);
	if (ret == 0)
		cntlr_send_1905_acknowledge(c, rx_cmdu, NULL, 0);

#ifdef CONTROLLER_SYNC_DYNAMIC_CNTLR_CONFIG
	if (proto == 0xac) {
		struct sync_config cfg = {0};
		uint16_t sync_config_respsize = 0;
		uint8_t *sync_config_resp = NULL;
		struct cmdu_buff *cmdu;

		dbg("******** Handle dyn-cntlr-sync-config-request ******\n");
		/* if (n->sync_config_req)
			free(n->sync_config_req);

		n->sync_config_reqsize = 0;
		n->sync_config_req = calloc(data_len, sizeof(uint8_t));
		if (n->sync_config_req) {
			memcpy(n->sync_config_req, data, data_len);
			n->sync_config_reqsize = data_len;
			fprintf(stderr, "%s: -- %d --\n", __func__, __LINE__);
		}
		*/
		ret = readfrom_configfile("/etc/config/mapcontroller", &cfg.data, &cfg.len);
		if (ret) {
			err("error reading controller config file\n");
			return -1;
		}

		ret = build_sync_config_response(data, data_len,
						 &cfg,
						 &sync_config_resp,
						 &sync_config_respsize);

		free(cfg.data);

		dbg("dyn-cntlr-sync-config response size = %zu\n", sync_config_respsize);

		/* ret = build_sync_config_response(c->sync_config_req,
						 c->sync_config_reqsize,
						 &cfg,
						 &sync_config_resp,
						 &sync_config_respsize); */
		if (ret) {
			dbg("Error building sync-config-response\n");
			return ret;
		}

		cmdu = cntlr_gen_higher_layer_data(c, origin, proto,
						   sync_config_resp,
						   sync_config_respsize);

		free(sync_config_resp);

		if (!cmdu)
			return -1;

		send_cmdu(c, cmdu);
		cmdu_free(cmdu);
	}
#endif

	return ret;
}

int handle_backhaul_sta_steer_response(void *cntlr, struct cmdu_buff *cmdu)
{
	/* TODO: implement */
	trace("%s: --->\n", __func__);
	return 0;
}

int handle_channel_scan_report(void *cntlr, struct cmdu_buff *cmdu)
{
	dbg("%s: --->\n", __func__);
	int i, j, num = 256;
	int offset = 0;
	uint8_t time_len;
	uint16_t num_neightbor;
	uint8_t bw_len;
	uint8_t *tv_data = NULL;
	struct tlv *tv_tsp[1][16];
	struct tlv *tv_scan[256];
	char timestamp[TIMESTAMP_MAX_LEN] = {0};
	struct tlv_timestamp *p = NULL;


	if (!validate_channel_scan_report(cmdu, tv_tsp, tv_scan, &num)) {
		dbg("cmdu validation: [CHANNEL_SCAN_REPORT] failed\n");
		return -1;
	}

	dbg("\nTLV type: MAP_TLV_TIMESTAMP\n");
	tv_data = (uint8_t *)tv_tsp[0][0]->data;
	p = (struct tlv_timestamp *)tv_data;

	dbg("\tlen: %d\n", p->len);

	/*Max size is 256 as per the Multi-ap r2 spec*/
	if (p->len > (TIMESTAMP_MAX_LEN - 1))
		return -1;

	strncpy(timestamp, (char *)p->timestamp, p->len);
	timestamp[p->len] = '\0';

	dbg("\ttimestamp: %s\n", timestamp);
	dbg("\n");

	for (i = 0; i < num; i++) {
		uint8_t info = 0x00;
		tv_data = (uint8_t *)tv_scan[i]->data;
		struct tlv_channel_scan_result *p1 = (struct tlv_channel_scan_result *)tv_data;

		dbg("\nTLV type: MAP_TLV_CHANNEL_SCAN_RES\n");
		dbg("\tradio: " MACFMT "\n", MAC2STR(p1->radio));
		dbg("\topclass: %d\n", p1->opclass);
		dbg("\tchannel: %d\n", p1->channel);
		dbg("\tstatus: 0x%02x\n", p1->status);

		offset = sizeof(*p1);
		if (p1->status == 0x00) {
			time_len = tv_data[offset++];

			dbg("\tlen: %d\n", time_len - 1);
			dbg("\ttimestamp: ");
			for (j = 0; j < time_len; j++) {
				dbg("%c", tv_data[offset]);
				offset++;
			}
			dbg("\n");

			dbg("\tutilization: %d\n", tv_data[offset]);
			offset++;
			dbg("\tnoise: %d\n", tv_data[offset]);
			offset++;
			num_neightbor = BUF_GET_BE16(tv_data[offset]);
			dbg("\tnum_neighbor: %d\n", num_neightbor);
			offset += 2;
			for (j = 0; j < num_neightbor; j++) {
				char ssidstr[33] = {0};
				uint8_t len = 0, ssidlen;

				dbg("\n\t\tbssid: " MACFMT "\n", MAC2STR(&tv_data[offset]));
				offset += 6;
				ssidlen = tv_data[offset++];
				dbg("\t\tlen: %d\n", ssidlen);
				len = (ssidlen + 1 > sizeof(ssidstr)
						? sizeof(ssidstr) : ssidlen + 1);
				snprintf(ssidstr, len, "%s", (char *)&tv_data[offset]);
				dbg("\t\tssid: %s\n", ssidstr);
				offset += ssidlen;
				dbg("\t\trcpi: %d\n", (int)tv_data[offset]);
				offset++;
				bw_len = tv_data[offset++];
				dbg("\t\tlen: %d\n", bw_len);
				dbg("\t\tbwstr: %d\n", atoi((char *)&tv_data[offset]));
				offset += bw_len;
				info = tv_data[offset];
				dbg("\t\tinfo: %d\n", offset);
				offset++;

				if (info & CH_SCAN_RESULT_BSSLOAD_PRESENT) {
					dbg("\t\t\tch_util: %d\n", tv_data[offset]);
					offset++;
					dbg("\t\t\tsta_count: %d\n", tv_data[offset]);
					offset += 2;
				}
			}
		}
		dbg("\n");
	}

	return 0;
}

int handle_sta_disassoc_stats(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);
	return 0;
}

int handle_assoc_status_notification(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);
	return 0;
}

int handle_tunneled_message(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);
	return 0;
}

int handle_backhaul_sta_caps_report(void *cntlr, struct cmdu_buff *cmdu)
{
	uint8_t *tv_data;
	struct tlv *tv[1][16];
	struct tlv_policy d_policy[] = {
		[0] = { .type = MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY, .present = TLV_PRESENT_NUM }
	};
	int num = 0;
	struct node *n;
	struct controller *c = (struct controller *) cntlr;

	cmdu_parse_tlvs(cmdu, tv, d_policy, 1);

	n = find_node_by_mac(c, cmdu->origin);
	if (!n)
		return -1;

	if (!tv[0][num]) {
		dbg("No TLV_BACKHAUL_STA_RADIO_CAPABILITY received!\n");
		return -1;
	}

	while (tv[0][num]) {
		struct netif_radio *r;
		struct tlv_bsta_radio_cap *p;

		if (tv[0][num]->type != MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY) {
			dbg("Wrong received TLV type!\n");
			return -1;
		}

		tv_data = (uint8_t *)tv[0][num]->data;
		p = (struct tlv_bsta_radio_cap *)tv_data;

		r = find_radio_by_node(c, n, p->radio);
		if (!r) {
			r = cntlr_node_add_radio(c, n, p->radio);
			if (!r)
				continue;
		}

		if (p->macaddr_included) {
			struct netif_iface *fh;

			fh = cntlr_radio_add_interface(c, r, (uint8_t *)p->macaddr);
			if (!fh)
				continue;

			fh->type = NETIF_BSTA;
		}

		num++;
	}
	return 0;
}


int handle_failed_connection_msg(void *cntlr, struct cmdu_buff *cmdu)
{
	trace("%s: --->\n", __func__);
	return 0;
}


//#define CMDU_TYPE_1905_START	0x0001
//#define CMDU_TYPE_1905_END	0x0009


#define CMDU_TYPE_MAP_START	0x8000
#define CMDU_TYPE_MAP_END	0x8033

/* mind the holes in the following two tables */
static const struct map_cmdu_calltable_t i1905ftable[] = {
	[0x01] = {
		.handle = handle_topology_notification,
		.debug = debug_topology_notification
	},
	[0x03] = {
		.handle = handle_topology_response,
		.debug = debug_topology_response
	},
	// [0x06] = {
	// 	.handle = handle_link_metric_response,
	// 	.debug = debug_link_metrics_response
	// },
	[0x07] = {
		.handle = handle_ap_autoconfig_search,
		.debug = debug_ap_autoconfig_search
	},
	[0x08] = {
		.handle = handle_ap_autoconfig_response,
		.debug = debug_ap_autoconfig_response
	},
	[0x09] = {
		.handle = handle_ap_autoconfig_wsc,
		.debug = debug_ap_autoconfig_wsc
	},
};


static const struct map_cmdu_calltable_t cntlr_mapftable[] = {
	[0x00] = {
		.handle = handle_1905_ack,
		.debug = debug_1905_ack
	},
	[0x02] = {
		.handle = handle_ap_caps_report,
		.debug = debug_ap_caps_report
	},
	[0x05] = {
		.handle = handle_channel_pref_report,
		.debug = debug_channel_pref_report
	},
	[0x07] = {
		.handle = handle_channel_sel_response,
		.debug = debug_channel_sel_response
	},
	[0x08] = {
		.handle = handle_oper_channel_report,
		.debug = debug_oper_channel_report
	},
	[0x0a] = {
		.handle = handle_sta_caps_report,
		.debug = debug_sta_caps_report
	},
	[0x0c] = {
		.handle = handle_ap_metrics_response,
		.debug = debug_ap_metrics_response
	},
	[0x0e] = {
		.handle = handle_sta_link_metrics_response,
		.debug = debug_sta_link_metrics_response
	},
	[0x10] = {
		.handle = handle_unassoc_sta_link_metrics_response,
		.debug = debug_unassoc_sta_link_metrics_response
	},
	[0x12] = {
		.handle = handle_beacon_metrics_response,
		.debug = debug_beacon_metrics_response
	},
	[0x15] = {
		.handle = handle_sta_steer_btm_report,
		.debug = debug_sta_steer_btm_report
	},
	[0x17] = {
		.handle = handle_sta_steer_complete,
		.debug = debug_sta_steer_complete
	},
	[0x18] = {
		.handle = handle_hld_message,
		.debug = debug_hld_message
	},
	[0x1a] = {
		.handle = handle_backhaul_sta_steer_response,
		.debug = debug_backhaul_sta_steer_response
	},
	[0x1c] = {
		.handle = handle_channel_scan_report,
		.debug = debug_channel_scan_report
	},
	[0x22] = {
		.handle = handle_sta_disassoc_stats,
		.debug = debug_sta_disassoc_stats
	},
	[0x25] = {
		.handle = handle_assoc_status_notification,
		.debug = debug_assoc_status_notification
	},
	[0x26] = {
		.handle = handle_tunneled_message,
		.debug = debug_tunneled_message
	},
	[0x28] = {
		.handle = handle_backhaul_sta_caps_report,
		.debug = debug_backhaul_sta_caps_report
	},
	[0x33] = {
		.handle = handle_failed_connection_msg,
		.debug = debug_failed_connection_msg
	},

};


bool is_cmdu_for_us(void *module, uint16_t type)
{
	struct controller *c = (struct controller *)module;

	/* TODO: handle cmdu types relevant for operating profile. */

	/* discard responses that are not to ACS during initilization */
	if (c->state == CNTLR_INIT &&
			(type != CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE &&
			 type != CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH)) {
		trace("Controller is initializing, waitfor ACS response or "\
				"start timer (%ds) to finish\n",
				uloop_timeout_remaining(&c->start_timer));
		return false;
	}

	if (type >= CMDU_TYPE_1905_START && type <= CMDU_TYPE_1905_END) {
		if (i1905ftable[type].handle)
			return true;
	} else if (type >= CMDU_TYPE_MAP_START && type <= CMDU_TYPE_MAP_END) {
		if (cntlr_mapftable[type - CMDU_TYPE_MAP_START].handle)
			return true;
	}

	return false;
}

int cntlr_handle_map_event(void *module, uint16_t cmdutype, uint16_t mid,
		char *rxif, uint8_t *src, uint8_t *origin, uint8_t *tlvs, int len)
{
	struct controller *c = (struct controller *)module;
	const struct map_cmdu_calltable_t *f;
	struct cmdu_buff *cmdu = NULL;
	int ret = 0;
	int idx;
	uint16_t resp_type;
	void *cookie;
	struct cmdu_ackq_entry *entry;

	trace("%s: ---> cmdu = %d\n", __func__, cmdutype);

	/* If request CMDU is from us, do not process is. This is for
	 * situation where controller and agent are on the same device,
	 * share the same MAC address and send CMDU's to each other. */
	if (hwaddr_equal(c->almac, origin)) {
		// Do we expect response for this CMDU ?
		resp_type = cntlr_cmdu_expect_response(c, cmdutype);
		entry = cmdu_ackq_lookup(&c->cmdu_ack_q, resp_type, mid, origin);
		if (entry)
			return 0;
	}

	ret = cmdu_ackq_dequeue(&c->cmdu_ack_q, cmdutype, mid, origin, &cookie);
	if (ret == 0)
		cmdu_free((struct cmdu_buff *) cookie);

	if (cmdutype >= CMDU_TYPE_MAP_START) {
		idx = cmdutype - CMDU_TYPE_MAP_START;
		f = cntlr_mapftable;
	} else {
		idx = cmdutype;
		f = i1905ftable;
	}

	cmdu = cmdu_alloc_custom(cmdutype, &mid, rxif, src, tlvs, len);
	if (!cmdu) {
		dbg("%s: cmdu_alloc_custom() failed!\n", __func__);
		return -1;
	}
	memcpy(cmdu->origin, origin, 6);
	dbg("%s: cmdu_alloc_custom() succeeded! cmdu->cdata->hdr.mid %u\n", __func__, cmdu_get_mid(cmdu));

	if (f[idx].handle)
		ret = f[idx].handle(c, cmdu);

	if (f[idx].debug)
		ret = f[idx].debug(c, cmdu);

	//TODO: check ret

	cmdu_free(cmdu);
	return ret;
}
